import bpy
from mathutils import Vector
from . import auto_rigger_settings

def create_fk_ik_controller_for_limbs():

    # Gets the bones pointer via Armature Prop as well as the edit bones
    reference = bpy.context.scene.ArmatureProp

    # Create IK limbs for hand,foot,knee, and elbow bones for left side
    create_fk_ik_controller_for_arm_limb('Arm FK/IK Controller', reference.Left_Lower_Arm, reference.Left_Hand, 'RootTransform', 'L')
    create_fk_ik_controller_for_arm_limb('Arm FK/IK Controller', reference.Right_Lower_Arm, reference.Right_Hand, 'RootTransform', 'R')
    create_fk_ik_controller_for_leg_limb(reference.Left_Knee, reference.Left_Foot, 'RootTransform', 'L')
    create_fk_ik_controller_for_leg_limb(reference.Right_Knee, reference.Right_Foot, 'RootTransform', 'R')


def create_fk_ik_controller_for_arm_limb(type_of_controller, main_bone, end_bone, parent_bone, s):
    # Sets up the inital variables

    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.transform_apply(rotation=True)
    bpy.ops.object.transform_apply(scale=True)
    bpy.ops.object.transform_apply(location=True)

    # Goes into edit mode mode and creates the FK/IK Controller above the arm_data bone
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones
    fk_ik_slider = edit_bones.new(name='Arm FK/IK Controller.' + s)
    fk_ik_slider.head = edit_bones[main_bone].head + Vector((0.0, 0.0, edit_bones[main_bone].length))
    fk_ik_slider.tail = fk_ik_slider.head + Vector((0.0, 0.0, edit_bones[main_bone].length / 3))
    con_length = edit_bones[main_bone].length / 2
    fk_ik_root = edit_bones.new(name=type_of_controller + ' Root.' + s)
    fk_ik_root.head = fk_ik_slider.head
    fk_ik_root.tail = fk_ik_slider.tail
    fk_ik_slider.parent = fk_ik_root
    fk_ik_root.parent = edit_bones[parent_bone]

    # Goes into pose mode
    bpy.ops.object.mode_set(mode='POSE')

    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones[type_of_controller + '.' + s]

    # Creates a limit location constraint
    constraints = selected_bone.constraints
    limit_con = constraints.new('LIMIT_LOCATION')
    limit_con.use_min_y = True
    limit_con.min_y = 0.00
    limit_con.use_max_y = True
    limit_con.max_y = 0.00
    limit_con.use_min_z = True
    limit_con.min_z = 0.00
    limit_con.use_max_z = True
    limit_con.max_z = 0.00
    limit_con.owner_space = 'LOCAL'
    limit_con.use_transform_limit = True
    limit_con.use_min_x = True

    # Sets the limits based on the "side" set.
    if s == 'L':
        limit_con.min_x = 0.0
        limit_con.use_max_x = True
        limit_con.max_x = con_length
    else:
        limit_con.min_x = -con_length
        limit_con.use_max_x = True
        limit_con.max_x = 0

    # Creates a driver to control the IK's Influence
    drv = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones[main_bone].constraints['IK'].driver_add('influence').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[auto_rigger_settings.Arm_name]
    target.bone_target = 'Arm FK/IK Controller.' + s
    target.transform_type = 'LOC_X'
    target.transform_space = 'LOCAL_SPACE'

    # Similar to how we set the limits to the limit location constraint
    if s == 'L':
        drv.expression = f"var*(1/{con_length:f})"
    else:
        drv.expression = f"var*(1/{-con_length:f})"

    # Creates a driver to control the influence of the rotation influence similar to the IK influence driver
    drv = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones[end_bone].constraints['Copy Rotation'].driver_add('influence').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[auto_rigger_settings.Arm_name]
    target.bone_target = 'Arm FK/IK Controller.' + s
    target.transform_type = 'LOC_X'
    target.transform_space = 'LOCAL_SPACE'
    if s == 'L':
        drv.expression = f"var*(1/{con_length:f})"
    else:
        drv.expression = f"var*(1/{-con_length:f})"

    # Puts the FK/IK bones in Bone Layer 2
    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + '.' + s].layers[1] = True
    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + '.' + s].layers[0] = False

    for i in range(2, 31):
        bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + '.' + s].layers[i] = False

    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + ' Root.' + s].layers[1] = True
    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + ' Root.' + s].layers[0] = False

    for i in range(2, 31):
        bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[type_of_controller + ' Root.' + s].layers[i] = False

    # Returns to Object mode
    bpy.ops.object.mode_set(mode='OBJECT')

def create_fk_ik_controller_for_leg_limb(main_bone, end_bone, parent_bone, s):
    # Sets up the inital variables
    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.transform_apply(rotation=True)
    bpy.ops.object.transform_apply(scale=True)
    bpy.ops.object.transform_apply(location=True)

    # Goes into edit mode mode and creates the FK/IK Controller above the arm_data bone
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones
    fk_to_ik_main_bone = edit_bones.new(name='Leg FK/IK Controller.' + s)
    if s == 'L':
        fk_to_ik_main_bone.head = edit_bones[main_bone].head + Vector(
            (edit_bones[main_bone].length, 0.0, 0.0))
    else:
        fk_to_ik_main_bone.head = edit_bones[main_bone].head + Vector(
            (-edit_bones[main_bone].length, 0.0, 0.0))
    fk_to_ik_main_bone.tail = fk_to_ik_main_bone.head + Vector(
        (0.0, 0.0, edit_bones[main_bone].length / 3))
    con_length = edit_bones[main_bone].length / 2
    fk_to_ik_main_bone_root = edit_bones.new(name='Leg FK/IK Controller Root.' + s)
    fk_to_ik_main_bone_root.head = fk_to_ik_main_bone.head
    fk_to_ik_main_bone_root.tail = fk_to_ik_main_bone.tail
    fk_to_ik_main_bone.parent = fk_to_ik_main_bone_root
    fk_to_ik_main_bone_root.parent = edit_bones[parent_bone]

    # Goes into pose mode
    bpy.ops.object.mode_set(mode='POSE')

    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones['Leg FK/IK Controller.' + s]

    # Creates a limit location constraint
    constraints = selected_bone.constraints
    limit_con = constraints.new('LIMIT_LOCATION')
    limit_con.use_min_y = True
    limit_con.min_y = 0.00
    limit_con.use_max_y = True
    limit_con.max_y = 0.00
    limit_con.owner_space = 'LOCAL'
    limit_con.use_transform_limit = True
    limit_con.use_min_x = True

    # Sets the limits based on the "side" set.
    if s == 'L':
        limit_con.min_x = 0.0
        limit_con.use_max_x = True
        limit_con.max_x = con_length
    else:
        limit_con.min_x = -con_length
        limit_con.use_max_x = True
        limit_con.max_x = 0

    # Creates a driver to control the IK's Influence
    drv = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones[main_bone].constraints['IK'].driver_add('influence').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[auto_rigger_settings.Arm_name]
    target.bone_target = 'Leg FK/IK Controller.' + s
    target.transform_type = 'LOC_X'
    target.transform_space = 'LOCAL_SPACE'

    # Similar to how we set the limits to the limit location constraint
    if s == 'L':
        drv.expression = f"1- (var*(1/{con_length:f}))"
    else:
        drv.expression = f"1- (var*(1/{-con_length:f}))"

    # Creates a driver to control the influence of the rotation influence similar to the IK influence driver
    drv = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones[end_bone].constraints['Copy Rotation'].driver_add('influence').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[auto_rigger_settings.Arm_name]
    target.bone_target = 'Leg FK/IK Controller.' + s
    target.transform_type = 'LOC_X'
    target.transform_space = 'LOCAL_SPACE'
    if s == 'L':
        drv.expression = f"1- (var*(1/{con_length:f}))"
    else:
        drv.expression = f"1- (var*(1/{-con_length:f}))"

    # Puts the FK/IK bones in Bone Layer 2

    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller.' + s].layers[1] = True
    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller.' + s].layers[0] = False

    for i in range(2, 31):
        bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller.' + s].layers[i] = False

    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller Root.' + s].layers[1] = True
    bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller Root.' + s].layers[0] = False

    for i in range(2, 31):
        bpy.data.objects[auto_rigger_settings.Arm_name].data.bones['Leg FK/IK Controller Root.' + s].layers[i] = False

    # Returns to Object mode
    bpy.ops.object.mode_set(mode='OBJECT')
