import bpy
from . import auto_rigger_settings
from mathutils import Vector, Matrix, Quaternion, Euler

# Auto Rigger IK Limbs: Responsible for creating and setting the IK limbs for the Auto Rigger


def signed_angle(vector_u, vector_v, normal):
    # Normal specifies orientation
    angle = vector_u.angle(vector_v)
    if vector_u.cross(vector_v).angle(normal) < 1:
        angle = -angle
    return angle


def get_pole_angle(base_bone, ik_bone, pole_location):
    pole_normal = (ik_bone.tail - base_bone.head).cross(pole_location - base_bone.head)
    projected_pole_axis = pole_normal.cross(base_bone.tail - base_bone.head)
    return signed_angle(base_bone.x_axis, projected_pole_axis, base_bone.tail - base_bone.head)


def create_hand_ik(edit_bones, hand_bone, side):

    # Stores the original hand bone and creates the new Hand IK Bone
    original_hand_bone = edit_bones[hand_bone]
    hand_ik_bone = edit_bones.new(name="HandIK." + side)

    # Sets the Hand IK in the same orientation as the original hand bone
    hand_ik_bone.head = original_hand_bone.head
    hand_ik_bone.tail = original_hand_bone.tail


def create_foot_ik(edit_bones, foot_bone, side):
    original_foot_bone = edit_bones[foot_bone]

    foot_ik_bone = edit_bones.new(name="FootIK." + side)
    foot_ik_bone.head = original_foot_bone.head
    foot_ik_bone.tail = original_foot_bone.tail


def create_elbow_ik(edit_bones, lowerarm, side):
    elbow_ik_bone = edit_bones.new(name="ElbowIK." + side)
    elbow_ik_bone.head = edit_bones[lowerarm].head + Vector((0.0, edit_bones[lowerarm].length, 0.0))
    elbow_ik_bone.tail = edit_bones[lowerarm].head + Vector((0.0, edit_bones[lowerarm].length, edit_bones[lowerarm].length / 3))


def create_knee_ik(edit_bones, knee, side):
    knee_ik_bone = edit_bones.new(name="KneeIK." + side)
    knee_ik_bone.head = edit_bones[knee].head + Vector((0.0, -(edit_bones[knee].length * 1.5), 0.0))
    knee_ik_bone.tail = edit_bones[knee].head + Vector((0.0, -(edit_bones[knee].length * 1.5), edit_bones[knee].length / 3))


def create_ik_limbs():

    # Gets the bones pointer via Armature Prop as well as the edit bones
    reference = bpy.context.scene.ArmatureProp
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones
    bpy.ops.object.mode_set(mode='EDIT')

    # Create IK limbs for hand,foot,knee, and elbow bones for left side
    create_hand_ik(edit_bones, reference.Left_Hand, "L")
    create_foot_ik(edit_bones, reference.Left_Foot, "L")
    create_elbow_ik(edit_bones, reference.Left_Lower_Arm, "L")
    create_knee_ik(edit_bones, reference.Left_Knee, "L")

    # Create IK limbs for hand,foot,knee, and elbow bones for right side
    create_hand_ik(edit_bones, reference.Right_Hand, "R")
    create_foot_ik(edit_bones, reference.Right_Foot, "R")
    create_elbow_ik(edit_bones, reference.Right_Lower_Arm, "R")
    create_knee_ik(edit_bones, reference.Right_Knee, "R")

    # Goes back to object mode
    bpy.ops.object.mode_set(mode='OBJECT')


def apply_ik_constraint(target_bone, ik_bone, pole_bone, pole_angle):

    IKcon = target_bone.constraints.new('IK')
    IKcon.target = bpy.data.objects[auto_rigger_settings.Arm_name]
    IKcon.subtarget = ik_bone.name
    IKcon.chain_count = 2
    IKcon.pole_target = bpy.data.objects[auto_rigger_settings.Arm_name]
    IKcon.pole_subtarget = pole_bone.name
    IKcon.pole_angle = pole_angle



def apply_copy_rotation_constraint(target_bone,ik_bone):

    Copycon = target_bone.constraints.new('COPY_ROTATION')
    Copycon.target = bpy.data.objects[auto_rigger_settings.Arm_name]
    Copycon.subtarget = ik_bone.name
    Copycon.target_space = 'WORLD'
    Copycon.owner_space = 'WORLD'



def apply_ik_constraint_to_limbs(selected_bone, rotation_bone, ik_bone, pole_ik_bone):
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones

    # Retrieves the pole angle for both the Arm and Leg bones
    bpy.ops.object.mode_set(mode='EDIT')
    arm_pole_angle = get_pole_angle(edit_bones[selected_bone], edit_bones[ik_bone], edit_bones[pole_ik_bone].matrix.translation)

    # Stores the pose bones wanting to be used
    bpy.ops.object.mode_set(mode='POSE')
    pose_bones = bpy.data.objects[auto_rigger_settings.Arm_name].pose.bones

    apply_ik_constraint(pose_bones[selected_bone], pose_bones[ik_bone], pose_bones[pole_ik_bone], arm_pole_angle)
    apply_copy_rotation_constraint(pose_bones[rotation_bone], pose_bones[ik_bone])
    fix_bone_roll(rotation_bone)


def fix_bone_roll(selected_bone):
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones

    print("Setting " + selected_bone + " roll to zero")
    selected_bone = edit_bones[selected_bone]

    selected_bone.roll = 0

def apply_ik_constraints():
    reference = bpy.context.scene.ArmatureProp

    apply_ik_constraint_to_limbs(reference.Left_Lower_Arm, reference.Left_Hand, 'HandIK.L', 'ElbowIK.L')
    apply_ik_constraint_to_limbs(reference.Right_Lower_Arm, reference.Right_Hand, 'HandIK.R', 'ElbowIK.R')
    apply_ik_constraint_to_limbs(reference.Left_Knee, reference.Left_Foot, 'FootIK.L', 'KneeIK.L')
    apply_ik_constraint_to_limbs(reference.Right_Knee, reference.Right_Foot, 'FootIK.R', 'KneeIK.R')


def create_root_transform(reference_bone):
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones

    # Creates the root transform and positions it at the center of the model
    bone = edit_bones.new(name='RootTransform')
    bone.head = Vector((0.0, 0.0, 0.0))
    destination = bone.head + Vector((0.0, -edit_bones[reference_bone].length, 0.0))
    bone.tail = destination


def connect_ik_bones_to_root_transform_bone():
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones
    reference = bpy.context.scene.ArmatureProp

    create_root_transform(reference.Left_Foot)
    ik_bones = ['HandIK.L', 'HandIK.R', 'FootIK.L', 'FootIK.R', 'ElbowIK.L', 'ElbowIK.R', 'KneeIK.L', 'KneeIK.R']

    for i in range(len(ik_bones)):
        edit_bones[ik_bones[i]].parent = edit_bones['RootTransform']

    edit_bones[reference.Pelvis].use_connect = False
    edit_bones[reference.Pelvis].parent = edit_bones['RootTransform']
    bpy.ops.object.mode_set(mode='OBJECT')




