import bpy
from . import rig_fixer
from . import  auto_rigger_validator
from . import auto_rigger_settings
from . import auto_rigger_ik_limbs
from . import auto_rigger_fk_ik_controls
from . import  auto_rigger_custom_shapes

# Auto Rigger V2 Panel: Responsible for running the Auto Rigger
class RunAutoRig(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "object.auto_rig"
    bl_label = "Run Auto Rig"


    def rig_the_armature(self):
        auto_rigger_ik_limbs.create_ik_limbs()
        auto_rigger_ik_limbs.apply_ik_constraints()
        auto_rigger_ik_limbs.connect_ik_bones_to_root_transform_bone()


    def validate_rig(self):
        if auto_rigger_validator.validate_all_limb_bones():
            return True
        else:
            return False


    def validate_fk_ik_controls(self):
        reference = bpy.context.scene.ArmatureProp

        if reference.UseFKIKControllers is True:
            return True
        else:
            return False


    def validate_custom_bone_shapes(self):
        reference = bpy.context.scene.ArmatureProp

        if reference.AddCustomBoneShapes is True:
            return True
        else:
            return False

    def clear(self):
        reference = bpy.context.scene.ArmatureProp
        prop_names = (prop.identifier for prop in reference.bl_rna.properties)
        for name in prop_names:
            reference.property_unset(name)

    def move_bones_to_correct_bone_layers(self):
        bpy.ops.object.mode_set(mode='POSE')

        # Gets the reference to all of the pose bones and the user selected options
        pose_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.bones
        reference = bpy.context.scene.ArmatureProp

        # Creates an array of bone names from prop group as well as the bones the auto rigger created
        prop_names = (prop.identifier for prop in reference.bl_rna.properties)
        bones_in_group = {getattr(reference, name) for name in prop_names}
        other_bones = ["HandIK.L","HandIK.R","FootIK.L","FootIK.R","ElbowIK.L","ElbowIK.R","KneeIK.L","KneeIK.R","RootTransform"]

        # Stores the base fingers that will be used to gather the rest of the finger bones
        base_finger_bones = [reference.Left_Base_Index_Finger, reference.Left_Base_Middle_Finger, reference.Left_Base_Ring_Finger,
                            reference.Left_Base_Pinky_Finger, reference.Left_Base_Thumb,reference.Right_Base_Index_Finger,
                             reference.Right_Base_Middle_Finger, reference.Right_Base_Ring_Finger,
                            reference.Right_Base_Pinky_Finger, reference.Right_Base_Thumb]

        finger_bones = list()


        # Gather the finger bones from the hand bones if No Fingers is not set
        if reference.NoFingers is False:
            for f in base_finger_bones:
                self.GatherFingerBones(bpy.data.objects[auto_rigger_settings.Arm_name].data.bones[f],finger_bones)

        # iterates through all of the pose bones checking to see if any of them showed up in the input.
        # If any of them did, then they go into bone layer 0, otherwise, all of them are moved to bone layer 7
        for p in pose_bones:
            bone_found = p.name in bones_in_group
            new_layers = [False] * 32

            if bone_found:
                new_layers[0] = True
            else:
                # Checks to the other bones
                other_bone_found = p.name in other_bones

                if other_bone_found:
                    new_layers[0] = True

                #If it's still not found in the other bones group, then it gets moved into the last bone layer
                else:
                    new_layers[7] = True

            p.layers = new_layers

        # After moving the main bones, the finger bone layers get set
        if reference.NoFingers is False:
            self.move_finger_bones_to_correct_bone_layers(finger_bones)



        bpy.ops.object.mode_set(mode='OBJECT')


    def move_finger_bones_to_correct_bone_layers(self,finger_bone_list):
        new_layers = [False] * 32
        new_layers[0] = True

        for k in finger_bone_list:
            k.layers = new_layers



    def move_fk_ik_controller_bones_to_correct_bone_layers(self):
        print("running")
        bpy.ops.object.mode_set(mode='POSE')

        # Creates an array for the IK and FK controllers
        ik_fk_controller_bones = ["Arm FK/IK Controller Root.L", "Arm FK/IK Controller.L", "Arm FK/IK Controller Root.R",
                                  "Arm FK/IK Controller.R", "Leg FK/IK Controller Root.R","Leg FK/IK Controller.R",
                                  "Leg FK/IK Controller Root.L","Leg FK/IK Controller.L"]

        pose_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.bones

        # If the user set to use FK To IK sliders, then it will move the FK to IK bones
        for p in pose_bones:
            print(p.name)
            bone_found = p.name in ik_fk_controller_bones
            new_layers = [False] * 32

            # If the bone was found, then it moves it to the second bone layer. Since the move bones ran already, it has to set the last bone layer to false
            # since it was already moved to it
            if bone_found:
                new_layers[1] = True
                new_layers[7] = False

            p.layers = new_layers

            bpy.ops.object.mode_set(mode='OBJECT')

    def GatherFingerBones(self,bone, bones_array):
        if len(bone.children) == 0:
            print(bone.name)
            bones_array.append(bone)
        else:
            self.GatherFingerBones(bone.children[0], bones_array)
            print(bone.name)
            bones_array.append(bone)

    @classmethod
    def poll(cls, context):
        return context.active_object is not None

    def execute(self, context):
        # Checks to make sure the user inputted the right information before rigging the armature
        if self.validate_rig():

            reference = bpy.context.scene.ArmatureProp

            # Some users don't want the rig to be cleaned, so this will determine whether to clean it or not
            if reference.CleanRig is True:
                rig_fixer.fix_rig()

            # Runs the necessary functions to rig the armature
            self.rig_the_armature()


            if self.validate_fk_ik_controls():
                auto_rigger_fk_ik_controls.create_fk_ik_controller_for_limbs()

            if self.validate_custom_bone_shapes():
                auto_rigger_custom_shapes.add_custom_bones()

            self.move_bones_to_correct_bone_layers()

            if self.validate_fk_ik_controls():
                self.move_fk_ik_controller_bones_to_correct_bone_layers()

            # Displays a sucess message when the armature has been rigged
            self.report({'INFO'}, 'Auto Rig sucessfully rigged ' + auto_rigger_settings.Arm_name + '!')

            # Clears the previously filled in data
            self.clear()
        else:
            self.report({'ERROR'}, 'Could not perform Auto Rig (Check for missing input parameters)')
        return {'FINISHED'}

