import bpy
from mathutils import Vector, Matrix, Quaternion, Euler
from . import auto_rigger_settings

# Rig Fixer.py: Contains a series of functions designed to make the rig safe enough for an IK Rig

# Applies the transform for the armature to make sure any changes made to the rig is proportional to the armature
def apply_transforms():
    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.transform_apply(rotation=True)
    bpy.ops.object.transform_apply(scale=True)
    bpy.ops.object.transform_apply(location=True)


# Gets the edit bones ready for editing
def prepare_edit_bones():

    # Applies the transforms to the armature
    apply_transforms()

    # Gets the edit bones
    bpy.ops.object.mode_set(mode='EDIT')
    edit_bones = bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones

    # Goes through all of the edit bones setting the roll bones to zero
    for i in bpy.data.objects[auto_rigger_settings.Arm_name].data.edit_bones:
        i.roll = 0.0

    # Returns the edit bones when done
    return edit_bones


# Connects and bend the arm bones so that the IK constraint can work
def fix_arm_bone(edit_bones, hand, elbow, arm):
    selected_hand_bone = edit_bones[hand]
    selected_hand_length = selected_hand_bone.head + Vector((selected_hand_bone.length, 0.0, 0.0))

    selected_hand_bone.tail = selected_hand_length
    selected_elbow_bone = edit_bones[elbow]

    selected_hand_bone.parent = selected_elbow_bone
    selected_elbow_bone.tail = selected_hand_bone.head

    selected_hand_bone.use_connect = True

    selected_arm_bone = edit_bones[arm]
    selected_elbow_bone.parent = selected_arm_bone

    selected_elbow_length = selected_elbow_bone.head + Vector((0.0, selected_elbow_bone.length / 12, 0.0))
    selected_elbow_bone.head = selected_elbow_length
    selected_arm_bone.tail = selected_elbow_bone.head


# Connects and bends the legs so the IK constraint can be used for the leg
def fix_leg_bone(edit_bones, foot, knee, thigh):
    selected_foot_bone = edit_bones[foot]
    selected_foot_bone_length = selected_foot_bone.head + Vector((0.0, -selected_foot_bone.length, 0.0))
    selected_foot_bone.tail = selected_foot_bone_length

    selected_knee_bone = edit_bones[knee]
    selected_foot_bone.parent = selected_knee_bone
    selected_knee_bone.tail = selected_foot_bone.head
    selected_foot_bone.use_connect = True

    selected_thigh_bone = edit_bones[thigh]
    selected_knee_bone.parent = selected_thigh_bone
    selected_knee_length = selected_knee_bone.head + Vector(
        (0.0, -selected_knee_bone.length / 12, selected_knee_bone.length / 12))

    selected_knee_bone.head = selected_knee_length
    selected_thigh_bone.tail = selected_knee_bone.head
    selected_knee_bone.use_connect = True


# Main function responsible for cleaning the rig
def fix_rig():

    # Gets the bones pointer via Armature Prop
    reference = bpy.context.scene.ArmatureProp

    # Prepares the rig for fixing and transitions into edit mode
    edit_bones = prepare_edit_bones()
    bpy.ops.object.mode_set(mode='EDIT')

    # Fix all limb bones
    fix_arm_bone(edit_bones, reference.Left_Hand, reference.Left_Lower_Arm, reference.Left_Upper_Arm)
    fix_arm_bone(edit_bones, reference.Right_Hand, reference.Right_Lower_Arm, reference.Right_Upper_Arm)
    fix_leg_bone(edit_bones, reference.Left_Foot, reference.Left_Knee, reference.Left_Thigh)
    fix_leg_bone(edit_bones, reference.Right_Foot, reference.Right_Knee, reference.Right_Thigh)

    # Exits out of Edit mode and goes into Object Mode
    bpy.ops.object.mode_set(mode='OBJECT')
    print("Rig preparation is done!")
